from jinja2 import Environment, PackageLoader, select_autoescape, FileSystemLoader
from masonite.exceptions import RequiredContainerBindingNotFound


def view(template='index', dictionary={}):
    env = Environment(
        loader=PackageLoader('resources', 'templates'),
        autoescape=select_autoescape(['html', 'xml'])
    )

    return env.get_template(template + '.html').render(dictionary)


class View:
    """
    Render template view
    """

    def __init__(self, container):
        self.dictionary = {}
        self.composers = {}
        self.container = container

        # If the cache_for method is declared
        self.cache = False
        # Cache time of cache_for
        self.cache_time = None
        # Cache type of cache_for
        self.cache_type = None

        self.template = None


    def render(self, template, dictionary={}):
        """
        Get the string contents of the view.
        """

        filename = template + '.html'

        if template.startswith('/'):
            location = template.split('/')
            filename = location[-1] + '.html'

            # If the location is more than 1 directory deep
            if len(location[1:-1]) > 1:
                loader = PackageLoader(*location[1:-1])
            else:
                loader = FileSystemLoader(str('/'.join(location[1:-1]) + '/'))
            
            self.env = Environment(
                loader=loader,
                autoescape=select_autoescape(['html', 'xml']),
                extensions=['jinja2.ext.loopcontrols']
            )
        else:
            self.env = Environment(
                loader=PackageLoader('resources', 'templates'),
                autoescape=select_autoescape(['html', 'xml']),
                extensions=['jinja2.ext.loopcontrols']
            )

        self.dictionary.update(dictionary)

        # Load template
        self.template = template

        # Check if use cache and return template from cache if exists
        if self.container.has('Cache') and self.__cached_template_exists() and not self.__is_expired_cache():
            return self.__get_cached_template()

        if template in self.composers:
            self.dictionary.update(self.composers[template])

        if '*' in self.composers:
            self.dictionary.update(self.composers['*'])

        self.rendered_template = self.env.get_template(filename).render(
            self.dictionary)

        return self

    def composer(self, composer_name, dictionary):
        """
        Updates composer dictionary
        """

        if isinstance(composer_name, str):
            self.composers[composer_name] = dictionary

        if isinstance(composer_name, list):
            for composer in composer_name:
                self.composers[composer] = dictionary

        return self

    def extend(self):
        pass

    def share(self, dictionary):
        """
        Updates the dictionary
        """

        self.dictionary.update(dictionary)
        return self

    def cache_for(self, time=None, type=None):
        """
        Set time and type for cache
        """

        if not self.container.has('Cache'):
            raise RequiredContainerBindingNotFound(
                "The 'Cache' container binding is required to use this method and wasn't found in the container. You may be missing a Service Provider"
            )

        self.cache = True
        self.cache_time = float(time)
        self.cache_type = type
        if self.__is_expired_cache():
            self.__create_cache_template(self.template)
        return self

    def __create_cache_template(self, template):
        """
        Save in the cache the template
        """

        self.container.make('Cache').store_for(
            template, self.rendered_template,
            self.cache_time, self.cache_type, '.html',
        )

    def __cached_template_exists(self):
        """
        Check if the cache template exists
        """

        return self.container.make('Cache').cache_exists(self.template)

    def __is_expired_cache(self):
        """
        Check if cache is expired
        """

        # Check if cache_for is set and configurate
        if self.cache_time is None or self.cache_type is None and self.cache:
            return True

        driver_cache = self.container.make('Cache')

        # True is expired
        return not driver_cache.is_valid(self.template)

    def __get_cached_template(self):
        """
        Return the rendered template
        """

        driver_cache = self.container.make('Cache')
        self.rendered_template = driver_cache.get(self.template)
        return self



class BaseDriver:
    _manager = None

    def driver(self, driver):
        return self._manager.driver(driver)
    
    def load_manager(self, manager):
        self._manager = manager
        return self


import smtplib
from email.message import EmailMessage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from masonite.contracts.MailContract import MailContract
from masonite.drivers.BaseMailDriver import BaseMailDriver


class MailSmtpDriver(BaseMailDriver, MailContract):
    """
    Mail smtp driver
    """

    def send(self, message_contents=None):
        config = self.config.DRIVERS['smtp']

        message = MIMEMultipart('alternative')

        if not message_contents:
            message_contents = self.message_body

        message_contents = MIMEText(message_contents, 'html')

        message['Subject'] = self.message_subject
        message['From'] = '{0} <{1}>'.format(self.config.FROM['name'], self.config.FROM['address'])
        message['To'] = self.to_address
        message.attach(message_contents)

        # Send the message via our own SMTP server.
        s = smtplib.SMTP('{0}:{1}'.format(config['host'], config['port']))
        s.login(config['username'], config['password'])

        # s.send_message(message)
        s.sendmail(self.config.FROM['name'], self.to_address, message.as_string())
        s.quit()



""" A View Service Provider """
from masonite.provider import ServiceProvider
from masonite.view import View


class ViewProvider(ServiceProvider):

    wsgi = False

    def register(self):
        view = View(self.app)
        self.app.bind('ViewClass', view)
        self.app.bind('View', view.render)

    def boot(self):
        pass



import os

from masonite.contracts.UploadContract import UploadContract
from masonite.drivers.BaseUploadDriver import BaseUploadDriver


class UploadDiskDriver(BaseUploadDriver, UploadContract):
    """
    Upload from the disk driver
    """

    def __init__(self, StorageConfig, Application):
        self.config = StorageConfig
        self.appconfig = Application

    def store(self, fileitem, location=None):
        filename = os.path.basename(fileitem.filename)

        # Check if is a valid extension
        self.validate_extension(filename)

        if not location:
            location = self.config.DRIVERS['disk']['location']

        location += '/'

        open(location + filename, 'wb').write(fileitem.file.read())

        return location + filename

    def store_prepend(self, fileitem, prepend, location=None):
        filename = os.path.basename(fileitem.filename)

        if not location:
            location = self.config.DRIVERS['disk']['location']

        open(location + prepend + filename, 'wb').write(fileitem.file.read())

        return location + prepend + filename



''' A Module Description '''
from masonite.facades.Auth import Auth

class LoginController(object):
    ''' Login Form Controller '''

    def __init__(self):
        pass

    def show(self, Request, Application):
        ''' Show the login page '''
        return view('auth/login', {'app': Application, 'Auth': Auth(Request)})

    def store(self, Request):
        if Auth(Request).login(Request.input('username'), Request.input('password')):
            Request.redirect('/home')
        else:
            Request.redirect('/login')
        return 'check terminal'

    def logout(self, Request):
        Auth(Request).logout()
        return Request.redirect('/login')



from abc import ABC, abstractmethod


class SessionContract(ABC):

    @abstractmethod
    def get(self): pass

    @abstractmethod
    def set(self): pass

    @abstractmethod
    def has(self): pass

    @abstractmethod
    def all(self): pass

    @abstractmethod
    def flash(self): pass

    @abstractmethod
    def reset(self): pass

    @abstractmethod
    def helper(self): pass



""" Module for the Routing System """
import cgi
import importlib
import json

from pydoc import locate

from config import middleware
from masonite.exceptions import RouteMiddlewareNotFound


class Route():
    """ Loads the environ """

    def __init__(self, environ=None):
        self.url_list = []

        if environ:
            self.environ = environ
            self.url = environ['PATH_INFO']

            if self.is_not_get_request():
                self.environ['QUERY_STRING'] = self.set_post_params()

    def load_environ(self, environ):
        self.environ = environ
        self.url = environ['PATH_INFO']

        if self.is_not_get_request():
            self.environ['QUERY_STRING'] = self.set_post_params()

        return self

    def get(self, route, output=None):
        """ Returns the output """
        return output

    def set_post_params(self):
        """ If the route is a Post, swap the QUERY_STRING """
        fields = None
        if self.is_not_get_request():
            if 'CONTENT_TYPE' in self.environ and 'application/json' in self.environ['CONTENT_TYPE']:
                try:
                    request_body_size = int(self.environ.get('CONTENT_LENGTH', 0))
                except (ValueError):
                    request_body_size = 0

                request_body = self.environ['wsgi.input'].read(request_body_size)
                return {'payload': json.loads(request_body)}
            else:
                fields = cgi.FieldStorage(
                    fp=self.environ['wsgi.input'], environ=self.environ, keep_blank_values=1)
                return fields

    def is_post(self):
        """ Check to see if the current request is a POST request """
        if self.environ['REQUEST_METHOD'] == 'POST':
            return True

        return False

    def is_not_get_request(self):
        if not self.environ['REQUEST_METHOD'] == 'GET':
            return True

        return False

    def compile_route_to_regex(self, route):
        # Split the route
        split_given_route = route.route_url.split('/')

        # compile the provided url into regex
        url_list = []
        regex = '^'
        for regex_route in split_given_route:
            if '@' in regex_route:
                if ':int' in regex_route:
                    regex += r'(\d+)'
                elif ':string' in regex_route:
                    regex += r'([a-zA-Z]+)'
                else:
                    # default
                    regex += r'(\w+)'
                regex += r'\/'

                # append the variable name passed @(variable):int to a list
                url_list.append(
                    regex_route.replace('@', '').replace(
                        ':int', '').replace(':string', '')
                )
            else:
                regex += regex_route + r'\/'

        self.url_list = url_list
        regex += '$'
        return regex

    def generated_url_list(self):
        return self.url_list


class BaseHttpRoute:
    method_type = 'GET'
    output = False
    route_url = None
    request = None
    named_route = None
    required_domain = None
    module_location = 'app.http.controllers'
    list_middleware = []

    def route(self, route, output):
        """ Loads the route into the class """

        # If the output specified is a string controller
        if isinstance(output, str):
            mod = output.split('@') 

            # Gets the controller name from the output parameter
            # This is used to add support for additional modules
            # like 'LoginController' and 'Auth.LoginController'
            get_controller = mod[0].split('.')[-1]

            # If trying to get an absolute path
            if mod[0].startswith('/'):
                self.module_location = '.'.join(mod[0].replace('/', '').split('.')[0:-1])

            try:
                # Import the module
                module = importlib.import_module(
                    '{0}.'.format(self.module_location) + get_controller)  
                # Get the controller from the module
                controller = getattr(module, get_controller)

                # Get the view from the controller
                view = getattr(controller(), mod[1])
                self.output = view
            except Exception as e:
                print('\033[93mWarning in routes/web.py!', e, '\033[0m')
        else:
            self.output = output
        self.route_url = route
        return self

    def domain(self, domain):
        self.required_domain = domain
        return self

    def module(self, module):
        self.module_location = module
        return self

    def has_required_domain(self):
        if self.request.has_subdomain() and (self.required_domain is '*' or self.request.subdomain == self.required_domain):
            return True
        return False

    def name(self, name):
        """ Specifies the name of the route """
        self.named_route = name
        return self

    def load_request(self, request):
        """ Load the request into this class """
        self.request = request
        return self

    def middleware(self, *args):
        """ Loads a list of middleware to run """
        self.list_middleware = args
        return self

    def run_middleware(self, type_of_middleware):
        """ type_of_middleware should be a string that contains either 'before' or 'after' """

        # Get the list of middleware to run for a route.
        for arg in self.list_middleware:

            # Locate the middleware based on the string specified
            try:
                located_middleware = self.request.app().resolve(locate(middleware.ROUTE_MIDDLEWARE[arg]))
            except KeyError:
                raise RouteMiddlewareNotFound("Could not find the '{0}' route middleware".format(arg))

            # If the middleware has the specific type of middleware
            # (before or after) then execute that
            if hasattr(located_middleware, type_of_middleware):
                getattr(located_middleware, type_of_middleware)()


class Get(BaseHttpRoute):
    """ Class for specifying GET requests """

    def __init__(self):
        self.method_type = 'GET'


class Post(BaseHttpRoute):
    """ Class for specifying POST requests """

    def __init__(self):
        self.method_type = 'POST'


class Put(BaseHttpRoute):

    def __init__(self):
        self.method_type = 'PUT'


class Patch(BaseHttpRoute):

    def __init__(self):
        self.method_type = 'PATCH'


class Delete(BaseHttpRoute):

    def __init__(self):
        self.method_type = 'DELETE'



"""
Request Module handles many different aspects of a single request
Methods which require the request and help ease development should
be put here.

Methods may return another object if necessary to expand capabilities
of this class.
"""
from urllib.parse import parse_qs
from http import cookies

import tldextract

from masonite.auth.Sign import Sign
from masonite.helpers.Extendable import Extendable
from masonite.helpers.time import cookie_expire_time
from cryptography.fernet import InvalidToken


class Request(Extendable):
    """
    Handles many different aspects of a single request.
    This is the object passed through to the controllers
    as a request paramter
    """

    def __init__(self, environ=None):
        self.cookies = []
        self._headers = []
        self.url_params = {}
        self.redirect_url = False
        self.redirect_route = False
        self.user_model = None
        self.subdomain = None
        self._status = '404 Not Found'

        if environ:
            self.load_environ(environ)

        self.encryption_key = False
        self.container = None

    def input(self, param):
        """
        Returns either the FORM_PARAMS during a POST request
        or QUERY_STRING during a GET request
        """

        # Special Request Methods
        if self.is_not_get_request():
            if isinstance(self.params, str):
                return parse_qs(self.params)[param][0]

            if isinstance(self.params, dict):
                return self.params[param]

            if not self.params[param].filename:
                return self.params[param].value

            if self.params[param].filename:
                return self.params[param]

        # GET Request Input
        if self.has(param):
            return parse_qs(self.params)[param][0]

        return False

    def is_post(self):
        if self.environ['REQUEST_METHOD'] == 'POST':
            return True

        return False

    def is_not_get_request(self):
        if not self.environ['REQUEST_METHOD'] == 'GET':
            return True

        return False

    def __set_request_method(self):
        if self.has('request_method'):
            self.environ['REQUEST_METHOD'] = self.input('request_method')
            return True

        return False

    def key(self, key):
        """
        Sets encryption key
        """

        self.encryption_key = key
        return self

    def all(self):
        """
        Returns all the params
        """

        if isinstance(self.params, str):
            return parse_qs(self.params)

        return self.params

    def load_app(self, app):
        self.container = app
        return self

    def load_environ(self, environ):
        self.environ = environ
        self.params = environ['QUERY_STRING']
        self.method = environ['REQUEST_METHOD']
        self.path = environ['PATH_INFO']

        if self.has('request_method'):
            self.__set_request_method()

        return self

    def app(self):
        return self.container

    def has(self, param):
        """
        Check if a param exists
        """

        if param in self.params:
            return True

        return False

    def status(self, status):
        self._status = status
        return self

    def get_status_code(self):
        return self._status

    def header(self, key, value=None, http_prefix=True):

        # Get Headers
        if value is None:
            if 'HTTP_{0}'.format(key) in self.environ:
                return self.environ['HTTP_{0}'.format(key)]
            elif key in self.environ:
                return self.environ[key]
            else:
                return None

        # Set Headers
        if http_prefix:
            self.environ['HTTP_{0}'.format(key)] = str(value)
            self._headers.append(('HTTP_{0}'.format(key), str(value)))
        else:
            self.environ[key] = str(value)
            self._headers.append((key, str(value)))
        return True

    def get_headers(self):
        return self._headers

    def reset_headers(self):
        self._headers = []

    def set_params(self, params):
        """
        Loads the params into the class.
        These parameters are where the developer can retrieve the
        /url/@variable:string/ from the url.
        """

        self.url_params.update(params)
        return self

    def param(self, parameter):
        """
        Retrieves the param from the URL.
        The "parameter" parameter in this method should be the name of the
        @variable passed into the url in web.py
        """

        if parameter in self.url_params:
            return self.url_params[parameter]
        return False

    def cookie(self, key, value, encrypt=True,
               http_only="HttpOnly;", path='/', expires=''):
        """
        Sets a cookie in the browser
        """

        if encrypt:
            value = Sign(self.encryption_key).sign(value)
        else:
            value = value

        if expires:
            expires = "Expires={0};".format(cookie_expire_time(expires))

        if not http_only:
            http_only = ""

        self.cookies.append(
            ('Set-Cookie', '{0}={1};{2} {3}Path={4}'.format(
                key, value, expires, http_only, path)))
        self.append_cookie(key, value)
        return self

    def get_cookies(self):
        """
        Retrieve all cookies from the browser
        """

        return self.cookies

    def get_cookie(self, provided_cookie, decrypt=True):
        """
        Retrieves a specific cookie from the browser
        """

        if 'HTTP_COOKIE' in self.environ:
            grab_cookie = cookies.SimpleCookie(self.environ['HTTP_COOKIE'])

            if provided_cookie in grab_cookie:
                if decrypt:
                    try:
                        return Sign(self.encryption_key).unsign(
                            grab_cookie[provided_cookie].value)
                    except InvalidToken:
                        self.delete_cookie(provided_cookie)
                        return None
                return grab_cookie[provided_cookie].value

        return None

    def append_cookie(self, key, value):
        if 'HTTP_COOKIE' in self.environ and self.environ['HTTP_COOKIE']:
            self.environ['HTTP_COOKIE'] += ';{0}={1}'.format(
                key, value)
        else:
            self.environ['HTTP_COOKIE'] = '{0}={1}'.format(
                key, value)

    def delete_cookie(self, key):
        """
        Delete cookie
        """
        self.cookie(key, '', expires='expired')

        if 'HTTP_COOKIE' in self.environ and self.environ['HTTP_COOKIE']:

            cookies = self.environ['HTTP_COOKIE'].split(';')
            for index, cookie in enumerate(cookies):
                if cookie.startswith(key):
                    # remove that cookie
                    del cookies[index]

            # put string back together
            self.environ['HTTP_COOKIE'] = ';'.join(cookies)
            return True
        return False

    def set_user(self, user_model):
        """
        Loads the user into the class
        """

        self.user_model = user_model
        return self

    def user(self):
        """
        Retreives the user model
        """

        return self.user_model

    def redirect(self, route, params={}):
        """
        Redirect the user based on the route specified
        """
        self.redirect_url = self.compile_route_to_url(route, params)
        return self

    def redirectTo(self, route_name, params={}):
        """
        Redirect to a named route
        """
        web_routes = self.container.make('WebRoutes')

        for route in web_routes:
            if route.named_route == route_name:
                self.redirect_url = self.compile_route_to_url(route.route_url, params)

        return self

    def reset_redirections(self):
        self.redirect_url = False
        self.redirect_route = False

    def back(self, input_parameter='back'):
        """
        Go to a named route with the back parameter
        """
        self.redirect(self.input(input_parameter))
        return self

    def compile_route_to_url(self, route, params={}):
        """
        Compile the route url into a usable url
        Converts /url/@id into /url/1. Used for redirection
        """

        if "http" in route:
            return route

        # Split the url into a list
        split_url = route.split('/')

        # Start beginning of the new compiled url
        compiled_url = '/'

        # Iterate over the list
        for url in split_url:
            if url:
                # if the url contains a parameter variable like @id:int
                if '@' in url:
                    url = url.replace('@', '').replace(
                        ':int', '').replace(':string', '')
                    compiled_url += str(params[url]) + '/'
                else:
                    compiled_url += url + '/'

        # The loop isn't perfect and may have an unwanted trailing slash
        if compiled_url.endswith('/') and not route.endswith('/'):
            compiled_url = compiled_url[:-1]

        # The loop isn't perfect and may have 2 slashes next to eachother
        if '//' in compiled_url:
            compiled_url = compiled_url.replace('//', '/')

        return compiled_url

    def has_subdomain(self):
        url = tldextract.extract(self.environ['HTTP_HOST'])

        if url.subdomain:
            self.subdomain = url.subdomain
            self.url_params.update({'subdomain': self.subdomain})
            return True

        return False

    def send(self, params):
        """
        With
        """

        self.set_params(params)
        return self

    def helper(self):
        return self



""" A RedirectionProvider Service Provider """
from masonite.provider import ServiceProvider
from masonite.drivers.QueueAsyncDriver import QueueAsyncDriver
from masonite.managers.QueueManager import QueueManager
from config import queue


class QueueProvider(ServiceProvider):

    wsgi = False

    def register(self):
        self.app.bind('QueueAsyncDriver', QueueAsyncDriver)
        self.app.bind('QueueManager', QueueManager)
        self.app.bind('QueueConfig', queue)

    def boot(self, QueueConfig, QueueManager):
        self.app.bind(
            'Queue',
            QueueManager(self.app).driver(QueueConfig.DRIVER)
        )



""" A Mail Service Provider """
from masonite.provider import ServiceProvider
from masonite.drivers.MailSmtpDriver import MailSmtpDriver
from masonite.drivers.MailMailgunDriver import MailMailgunDriver
from masonite.managers.MailManager import MailManager

from config import mail


class MailProvider(ServiceProvider):

    wsgi = False

    def register(self):
        self.app.bind('MailConfig', mail)
        self.app.bind('MailSmtpDriver', MailSmtpDriver)
        self.app.bind('MailMailgunDriver', MailMailgunDriver)
        self.app.bind('MailManager', MailManager(self.app))

    def boot(self, MailConfig, MailManager):
        self.app.bind('Mail', MailManager.driver(MailConfig.DRIVER))



import os
from cleo import Command


class ProviderCommand(Command):
    """
    Creates a new Service Provider

    provider
        {name : Name of the Service Provider you want to create}
    """

    def handle(self):
        provider = self.argument('name')

        if not os.path.isfile('app/providers/{0}.py'.format(provider)):
            if not os.path.exists(os.path.dirname('app/providers/{0}.py'.format(provider))):
                # Create the path to the service provider if it does not exist
                os.makedirs(os.path.dirname('app/providers/{0}.py'.format(provider)))

            f = open('app/providers/{0}.py'.format(provider), 'w+')

            f.write("''' A {0} Service Provider '''\n".format(provider))
            f.write('from masonite.provider import ServiceProvider\n\n')
            f.write("class {0}(ServiceProvider):\n\n    ".format(provider))
            f.write("def register(self):\n        pass\n\n    ")
            f.write("def boot(self):\n        pass\n")

            self.info('Service Provider Created Successfully!')
        else:
            self.comment('Service Provider Already Exists!')



from cleo import Command


class MigrateRefreshCommand(Command):
    """
    Migrate refresh

    migrate:refresh
    """

    def handle(self):
        self.call('migrate:reset')
        self.call('migrate')



""" A RouteProvider Service Provider """
import re
import json
from pydoc import locate

from masonite.provider import ServiceProvider
from masonite.view import View


class RouteProvider(ServiceProvider):

    def register(self):
        pass

    def boot(self, WebRoutes, Route, Request, Environ, Headers):
        # All routes joined
        for route in WebRoutes:
            router = Route
            request = Request

            # Compiles the given route to regex
            regex = router.compile_route_to_regex(route)

            """
            |--------------------------------------------------------------------------
            | Make a better match for trailing slashes
            |--------------------------------------------------------------------------
            |
            | Sometimes a user will end with a trailing slash. Because the user might
            | create routes like `/url/route` and `/url/route/` and how the regex
            | is compiled down, we may need to adjust for urls that end or dont
            | end with a trailing slash.
            |
            """

            if route.route_url.endswith('/'):
                matchurl = re.compile(regex.replace(r'\/\/$', r'\/$'))
            else:
                matchurl = re.compile(regex.replace(r'\/$', r'$'))

            # This will create a dictionary of parameters given.
            # This is sort of a short
            # but complex way to retrieve the url parameters.
            # This is the code used to
            # convert /url/@firstname/@lastname to
            # {'firstmane': 'joseph', 'lastname': 'mancuso'}
            try:
                parameter_dict = {}
                for index, value in enumerate(matchurl.match(router.url).groups()):
                    parameter_dict[router.generated_url_list()[index]] = value
                request.set_params(parameter_dict)
            except AttributeError:
                pass

            """
            |--------------------------------------------------------------------------
            | Houston, we've got a match
            |--------------------------------------------------------------------------
            |
            | Check to see if a route matches the corresponding router url. If a match
            | is found, execute that route and break out of the loop. We only need
            | one match. Routes are executed on a first come, first serve basis
            |
            """

            if matchurl.match(router.url) and route.method_type == Environ['REQUEST_METHOD']:
                route.load_request(request)
                if request.has_subdomain():
                    # check if the subdomain matches the routes domain
                    if not route.has_required_domain():
                        self.app.bind('Response', 'Route not found. Error 404')
                        break
                """
                |--------------------------------------------------------------------------
                | Execute Before Middleware
                |--------------------------------------------------------------------------
                |
                | This is middleware that contains a before method.
                |
                """

                for http_middleware in self.app.make('HttpMiddleware'):
                    located_middleware = self.app.resolve(
                        locate(http_middleware)
                    )
                    if hasattr(located_middleware, 'before'):
                        located_middleware.before()

                # Show a helper in the terminal of which route has been visited
                print(route.method_type + ' Route: ' + router.url)

                # Loads the request in so the middleware
                # specified is able to use the
                # request object. This is before middleware and is ran
                # before the request
                route.run_middleware('before')

                # Get the data from the route. This data is typically the
                # output of the controller method
                if not request.redirect_url:
                    Request.status('200 OK')

                    response = self.app.resolve(route.output)

                    if isinstance(response, View):
                        response = response.rendered_template

                    self.app.bind(
                        'Response',
                        router.get(route.route, response)
                    )

                    # If the Content-Type was not set in the view or before this
                    if not Request.header('Content-Type'):
                        if isinstance(response, dict):
                            Request.header('Content-Type', 'application/json; charset=utf-8', http_prefix=None)
                            self.app.bind(
                                'Response',
                                str(json.dumps(response))
                            )
                        else:
                            Request.header('Content-Type', 'text/html; charset=utf-8', http_prefix=None)

                # Loads the request in so the middleware
                # specified is able to use the
                # request object. This is after middleware
                # and is ran after the request
                route.run_middleware('after')

                """
                |--------------------------------------------------------------------------
                | Execute After Middleware
                |--------------------------------------------------------------------------
                |
                | This is middleware with an after method.
                |
                """

                for http_middleware in self.app.make('HttpMiddleware'):
                    located_middleware = self.app.resolve(
                        locate(http_middleware)
                    )
                    if hasattr(located_middleware, 'after'):
                        located_middleware.after()
                break
            else:
                self.app.bind('Response', 'Route not found. Error 404')



""" An Upload Service Provider """
from masonite.provider import ServiceProvider
from masonite.managers.UploadManager import UploadManager
from masonite.drivers.UploadDiskDriver import UploadDiskDriver
from masonite.drivers.UploadS3Driver import UploadS3Driver
from config import storage


class UploadProvider(ServiceProvider):

    wsgi = False

    def register(self):
        self.app.bind('StorageConfig', storage)
        self.app.bind('UploadDiskDriver', UploadDiskDriver)
        self.app.bind('UploadS3Driver', UploadS3Driver)
        self.app.bind('UploadManager', UploadManager(self.app))

    def boot(self, UploadManager, StorageConfig):
        self.app.bind('Upload', UploadManager.driver(StorageConfig.DRIVER))



""" A SassProvider Service Provider """
import builtins
import os

from masonite.provider import ServiceProvider
from masonite.helpers.view_helpers import set_request_method


class HelpersProvider(ServiceProvider):

    wsgi = False

    def register(self):
        pass

    def boot(self, View, ViewClass, Request):
        """ Add helper functions to Masonite """
        builtins.view = View
        builtins.request = Request.helper
        builtins.auth = Request.user
        builtins.container = self.app.helper
        builtins.env = os.getenv
        builtins.resolve = self.app.resolve

        ViewClass.share(
            {
                'request': Request.helper,
                'auth': Request.user,
                'request_method': set_request_method
            }
        )



from masonite.view import view
from masonite.drivers.BaseDriver import BaseDriver


class BaseMailDriver(BaseDriver):
    """
    Class base for mail drivers
    """

    def __init__(self, MailConfig):
        self.config = MailConfig
        self.to_address = None
        self.from_address = self.config.FROM
        self.message_subject = 'Subject'
        self.message_body = None

    def to(self, user_email):
        if callable(user_email):
            user_email = user_email.email

        self.to_address = user_email
        return self

    def template(self, template_name, dictionary={}):
        self.message_body = view(template_name, dictionary)
        return self

    def send_from(self, address):
        self.from_address = address
        return self

    def subject(self, subject):
        self.message_subject = subject
        return self



from cleo import Command
import os


class ControllerCommand(Command):
    """
    Creates a controller

    controller
        {name : Name of the controller you would like to create}
        {--r|--resource : Create a controller as a resource}
        {--e|--exact : For add the name controller without `Controller` text}
    """

    def handle(self):
        controller = self.argument('name')

        if not self.option('exact'):
            controller = controller + "Controller"

        if os.path.isfile('app/http/controllers/{0}.py'.format(controller)):
            self.error('{0} Controller Exists!'.format(controller))
        else:
            f = open('app/http/controllers/{0}.py'.format(controller), 'w+')
            f.write("''' A Module Description '''\n\n")
            f.write('class {0}:\n'.format(controller))
            f.write("    ''' Class Docstring Description '''\n\n")
            f.write('    def show(self):\n')
            f.write('        pass\n')

            if self.option('resource'):
                f.write('\n    def index(self):\n')
                f.write('        pass\n\n')
                f.write('    def create(self):\n')
                f.write('        pass\n\n')
                f.write('    def store(self):\n')
                f.write('        pass\n\n')
                f.write('    def edit(self):\n')
                f.write('        pass\n\n')
                f.write('    def update(self):\n')
                f.write('        pass\n\n')
                f.write('    def destroy(self):\n')
                f.write('        pass\n')

            self.info('{0} Created Successfully!'.format(controller))



""" An ApiProvider Service Provider """
from masonite.provider import ServiceProvider


class ApiProvider(ServiceProvider):

    def register(self):
        pass

    def boot(self, Response, ApiRoutes, Route, Request):
        router = Route
        if Response == 'Route not found. Error 404':

            """
            |--------------------------------------------------------------------------
            | Pull in the API Routes from the Service Container
            |--------------------------------------------------------------------------
            |
            | The Service Container has loaded all Api routes into the container so
            | let's loop through and check for any matches.
            |
            """

            for route in ApiRoutes:

                """
                |--------------------------------------------------------------------------
                | If We've Got A Match
                |--------------------------------------------------------------------------
                |
                | If we have a match then let's go ahead and execute the route, load the 
                | response into the data variable and get on with our lives.
                |
                """

                if route.url in router.url:
                    data = route.fetch(Request).output
                    if data:
                        break
                    else:
                        data = 'Route not found. Error 404'
                else:
                    data = 'Route not found. Error 404'



class DriverNotFound(Exception):
    pass


class DriverLibraryNotFound(Exception):
    pass


class FileTypeException(Exception):
    """
    For exceptions extension invalid
    """
    pass


class RequiredContainerBindingNotFound(Exception):
    pass


class MissingContainerBindingNotFound(Exception):
    pass


class ContainerError(Exception):
    pass


class InvalidCSRFToken(Exception):
    """
    For exceptions that return error when verifying the csrf token
    """
    pass

class RouteMiddlewareNotFound(Exception): pass


from masonite.contracts.BroadcastContract import BroadcastContract
from masonite.exceptions import DriverLibraryNotFound
from masonite.drivers.BaseDriver import BaseDriver


class BroadcastAblyDriver(BroadcastContract, BaseDriver):

    def __init__(self, BroadcastConfig):
        self.config = BroadcastConfig
        self.ssl_message = True

    def ssl(self, boolean):
        self.ssl_message = boolean
        return self

    def channel(self, channels, message, event='base-event'):
        try:
            from ably import AblyRest
        except ImportError:
            raise DriverLibraryNotFound(
                'Could not find the "ably" library. Please pip install this library running "pip install ably"')

        client = AblyRest('{0}'.format(
            self.config.DRIVERS['ably']['secret']
        ))

        if isinstance(channels, list):
            for channel in channels:
                ably_channel = client.channels.get(channel)
                ably_channel.publish(event, message)
        else:
            channel = client.channels.get(channels)
            channel.publish(event, message)

        return message



import os
from cleo import Command


class JobCommand(Command):
    """
    Creates a new Job

    job
        {name : Name of the job you want to create}
    """

    def handle(self):
        job = self.argument('name')
        if not os.path.isfile('app/jobs/{0}.py'.format(job)):
            if not os.path.exists(os.path.dirname('app/jobs/{0}.py'.format(job))):
                # Create the path to the job if it does not exist
                os.makedirs(os.path.dirname('app/jobs/{0}.py'.format(job)))

            f = open('app/jobs/{0}.py'.format(job), 'w+')

            f.write("''' A {0} Queue Job '''\n\n".format(job))
            f.write('from masonite.queues.Queueable import Queueable\n\n')
            f.write("class {0}(Queueable):\n\n    ".format(job))
            f.write("def __init__(self):\n        pass\n\n    ")
            f.write("def handle(self):\n        pass\n")

            self.info('Job Created Successfully!')
        else:
            self.comment('Job Already Exists!')



''' A RedirectionProvider Service Provider '''
from masonite.provider import ServiceProvider
from masonite.managers.BroadcastManager import BroadcastManager
from masonite.drivers.BroadcastPusherDriver import BroadcastPusherDriver
from masonite.drivers.BroadcastAblyDriver import BroadcastAblyDriver
from config import broadcast


class BroadcastProvider(ServiceProvider):

    wsgi = False

    def register(self):
        self.app.bind('BroadcastConfig', broadcast)
        self.app.bind('BroadcastPusherDriver', BroadcastPusherDriver)
        self.app.bind('BroadcastAblyDriver', BroadcastAblyDriver)
        self.app.bind('BroadcastManager', BroadcastManager(self.app))

    def boot(self, BroadcastConfig, BroadcastManager):
        self.app.bind('Broadcast', self.app.make('BroadcastManager').driver(BroadcastConfig.DRIVER))



import threading

from masonite.contracts.QueueContract import QueueContract
from masonite.drivers.BaseDriver import BaseDriver


class QueueAsyncDriver(QueueContract, BaseDriver):
    """
    Queue Aysnc Driver
    """
    def __init__(self, Container):
        self.container = Container

    def push(self, *objects):
        for obj in objects:
            obj = self.container.resolve(obj)
            thread = threading.Thread(target=obj.dispatch(), args=(), kwargs={})
            thread.start()



import os
from cleo import Command


class CommandCommand(Command):
    """
    Creates a new command

    command
        {name : Name of the command you would like to create}
    """

    def handle(self):
        command = self.argument('name')
        if not os.path.isfile('app/commands/{0}.py'.format(command)):
            if not os.path.exists(os.path.dirname('app/commands/{0}.py'.format(command))):
                # Create the path to the command if it does not exist
                os.makedirs(os.path.dirname('app/commands/{0}.py'.format(command)))

            f = open('app/commands/{0}.py'.format(command), 'w+')

            f.write('""" A {0} Command """\n'.format(command))
            f.write('from cleo import Command\n\n\n')
            f.write('class {0}(Command):\n    """\n    Description of command\n\n    '.format(command))
            f.write('command:name\n        {argument : description}\n    """\n\n    ')
            f.write('def handle(self):\n        pass')

            self.info('Command Created Successfully!')
        else:
            self.error('Command Already Exists!')



class ServiceProvider():

    wsgi = True

    def __init__(self):
        self.app = None

    def boot(self):
        pass

    def register(self):
        self.app.bind('Request', object)

    def load_app(self, app):
        self.app = app
        return self



""" Masonite Validator Module """

from validator import *


class Validator:
    """ 
    Validator Class 
    """

    def __init__(self, request=None):
        self.request = request
        self.validation_dictionary = {}
        self.check_manual_dictionary = False
        self.error_messages = {}

    def validate(self, dictionary):
        """ Sets the validation dictionary """
        self.validation_dictionary = dictionary
        return self

    def check(self, check_manual_dictionary=False):
        """ Validated the dictionary """
        if check_manual_dictionary:
            self.check_manual_dictionary = check_manual_dictionary
        return self.run_validation()[0]

    def errors(self):
        """ Returns Errors """
        validation = self.run_validation()
        for message in self.error_messages:
            if message in validation[1]:
                validation[1][message] = self.error_messages[message]

        if not validation[1]:
            return None

        return validation[1]

    def run_validation(self):
        """ Loads the dictionary and runs the validations """
        self.load_request_input()
        return validate(self.validation_dictionary, self.load_request_input())

    def load_request_input(self):
        """ Need to load request input into a different value
            Since the Request class stores input values as string, we
                need to create a new dictonary value from the request
        """
        if self.request:
            dictionary = {}
            for value in self.request.all():
                dictionary[value] = self.request.input(value)
            return dictionary

        return self.check_manual_dictionary

    def messages(self, messages):
        """ You may change the default messages here """
        self.error_messages = messages



from subprocess import call
from cleo import Command


class ServeCommand(Command):
    """
    Run the Masonite server

    serve
        {--port=8000 : Specify which port to run the server}
        {--host=127.0.0.1 : Specify which ip address to run the server}
    """

    def handle(self):
        try:
            call([
                "waitress-serve", '--port', self.option('port'),
                "--host", self.option('host'), "wsgi:application"
            ])
        except Exception:
            self.line('')
            self.comment('Server aborted!')



""" Third party package integrations """
import os
import shutil
import sys


def create_or_append_config(location, name=False):
    if name:
        file_name = name
    else:
        file_name = os.path.basename(location)

    # import it into the config directory
    config_directory = os.path.join(os.getcwd(), 'config')

    # if file does not exist
    if not os.path.isfile(config_directory + '/' + file_name):
        shutil.copyfile(location,
                        config_directory + '/' + file_name)
        print('\033[92mConfiguration File Created!\033[0m')
    else:
        # Append to the file
        # tempfiles is a list of file handles to your temp files.
        project_config = open(config_directory + '/' + file_name, "a")
        package_config = open(location, 'r')

        project_config.write(package_config.read())

        project_config.close()
        package_config.close()
        print('\033[92mConfiguration File Appended!\033[0m')


def append_web_routes(location):
    # import it into the web.py file
    routes_file = os.path.join(os.getcwd(), 'routes/web.py')

    # Append to the file
    # tempfiles is a list of file handles to your temp files.
    project_routes = open(routes_file, "a")
    package_routes = open(location, 'r')

    project_routes.write(package_routes.read())

    project_routes.close()
    package_routes.close()
    print('\033[92mroutes/web.py File Appended!\033[0m')


def append_api_routes(location):
    # import it into the web.py file
    api_file = os.path.join(os.getcwd(), 'routes/api.py')

    # Append to the file
    # tempfiles is a list of file handles to your temp files.
    project_routes = open(api_file, "a")
    package_routes = open(location, 'r')

    project_routes.write(package_routes.read())

    project_routes.close()
    package_routes.close()
    print('\033[92mroutes/api.py File Appended!\033[0m')


def create_controller(location, to='app/http/controllers'):
    file_name = os.path.basename(location)

    controller_directory = os.path.join(os.getcwd(), to)
    controller_file = os.path.join(controller_directory, file_name)
    if not os.path.exists(controller_directory):
        # Create the path to the model if it does not exist
        os.makedirs(controller_directory)

    if os.path.isfile(controller_file):
        # if file does exist
        print('\033[91m{0} Controller Already Exists!\033[0m'.format(file_name))
    else:
        # copy controller over
        shutil.copyfile(
            location,
            controller_file
        )

        print('\033[92m{0} Controller Created\033[0m'.format(file_name))


def add_venv_site_packages():
    try:
        from config import packages
        # Add additional site packages to vendor if they exist
        for directory in packages.SITE_PACKAGES:
            path = os.path.join(os.getcwd(), directory)
            sys.path.append(path)
    except ImportError:
        raise ImportError

    if 'VIRTUAL_ENV' in os.environ:
        python_version = None
        venv_directory = os.listdir(
            os.path.join(os.environ['VIRTUAL_ENV'], 'lib')
        )

        for directory in venv_directory:
            if directory.startswith('python'):
                python_version = directory
                break

        if python_version:
            site_packages_directory = os.path.join(
                os.environ['VIRTUAL_ENV'],
                'lib',
                python_version,
                'site-packages'
            )

            sys.path.append(site_packages_directory)
        else:
            print('\033[93mWARNING: Could not add the virtual environment you are currently in. Attempting to add: {0}\033[93m'.format(
                os.environ['VIRTUAL_ENV']))



import glob
import os
import time

from masonite.contracts.CacheContract import CacheContract
from masonite.drivers.BaseDriver import BaseDriver


class CacheDiskDriver(CacheContract, BaseDriver):
    """
    Cache from the disk driver
    """

    def __init__(self, CacheConfig, Application):
        self.config = CacheConfig
        self.appconfig = Application
        self.cache_forever = None

    def store(self, key, value, extension=".txt", location=None):
        """
        Store content in cache file
        """

        self.cache_forever = True
        if not location:
            location = self.config.DRIVERS['disk']['location']

        location += '/'
        path = os.path.join(location, key + extension)
        if not os.path.exists(path):
            self._create_directory(path)

        open(path, 'w').write(value)

        return key

    def store_for(self, key, value, cache_time, cache_type, extension=".txt"):
        """
        Store content with time, type and extension
        """

        self.cache_forever = False
        cache_type = cache_type.lower()
        calc = 0

        if cache_type in ("second", "seconds"):
            # Set time now for
            calc = 1
        elif cache_type in ("minute", "minutes"):
            calc = 60
        elif cache_type in ("hour", "hours"):
            calc = 60 * 60
        elif cache_type in ("day", "days"):
            calc = 60 * 60 * 60
        elif cache_type in ("month", "months"):
            calc = 60 * 60 * 60 * 60
        elif cache_type in ("year", "years"):
            calc = 60 * 60 * 60 * 60 * 60
        else:
            raise ValueError(
                '{0} is not a valid caching type.'.format(cache_type))

        cache_for_time = cache_time * calc

        cache_for_time = cache_for_time + time.time()

        key = self.store(
            key + ":" + str(cache_for_time),
            value, extension,
        )

        return key

    def get(self, key):
        """
        Get the data from a key in the cache
        """

        if not self.is_valid(key):
            return None

        cache_path = self.config.DRIVERS['disk']['location'] + "/"
        content = ""

        if self.cache_forever:
            glob_path = cache_path + key + '*'
        else:
            glob_path = cache_path + key + ':*'

        try:
            content = open(glob.glob(glob_path)[0], 'r').read()
        except IndexError:
            pass

        return content

    def delete(self, key):
        """
        Delete file cache
        """

        cache_path = self.config.DRIVERS['disk']['location'] + "/"
        if self.cache_forever:
            glob_path = cache_path + key + '*'
        else:
            glob_path = cache_path + key + ':*'

        for template in glob.glob(glob_path):
            os.remove(template)

    def update(self, key, value, location=None):
        """
            Updates a specific cache by key
        """

        if not location:
            location = self.config.DRIVERS['disk']['location'] + "/"

        location = os.path.join(location, key)
        cache = glob.glob(location + ':*')[0]

        open(cache, 'w').write(str(value))

        return key

    def cache_exists(self, key):
        """
        Check if the cache exists
        """

        cache_path = self.config.DRIVERS['disk']['location'] + "/"
        if self.cache_forever:
            glob_path = cache_path + key + '*'
        else:
            glob_path = cache_path + key + ':*'

        find_template = glob.glob(glob_path)
        if find_template:
            return True
        return False

    def is_valid(self, key):
        """
        Check if a valid cache
        """

        cache_path = self.config.DRIVERS['disk']['location'] + "/"
        if self.cache_forever:
            glob_path = cache_path + key + '*'
        else:
            glob_path = cache_path + key + ':*'

        cache_file = glob.glob(glob_path)
        if cache_file:
            try:
                cache_timestamp = float(
                    os.path.splitext(cache_file[0])[0].split(':')[1]
                )
            except IndexError:
                if self.cache_forever:
                    return True
                else:
                    return False

            if cache_timestamp > time.time():
                return True

        self.delete(key)
        return False

    def _create_directory(self, directory):
        if not os.path.exists(os.path.dirname(directory)):
            # Create the path to the model if it does not exist
            os.makedirs(os.path.dirname(directory))
            return True
        return False



import inspect

class Extendable:

    def extend(self, key, obj=None):
        
        # If both key and an object is passed
        if obj:
            if inspect.ismethod(obj):
                obj = obj.__func__

            setattr(self, key, obj.__get__(self))
            return self

        # Extend all of a classes methods into this class
        if inspect.isclass(key):
            for method in inspect.getmembers(key, inspect.isfunction):
                setattr(self, method[0], method[1].__get__(self))

        # Extend a function into this class
        elif inspect.isfunction(key):
            setattr(self, key.__name__, key.__get__(self))
        elif inspect.ismethod(obj):
            setattr(self, key.__name__, key.__func__.__get__(self))
        return self



from abc import ABC, abstractmethod


class UploadContract(ABC):

    @abstractmethod
    def accept(self): pass

    @abstractmethod
    def validate_extension(self): pass

    @abstractmethod
    def store(self): pass

    @abstractmethod
    def store_prepend(self): pass



import requests
from masonite.contracts.MailContract import MailContract
from masonite.drivers.BaseMailDriver import BaseMailDriver
from masonite.drivers.BaseDriver import BaseDriver


class MailMailgunDriver(BaseMailDriver, MailContract):
    """
    Mailgun driver
    """
    def send(self, message=None):
        if not message:
            message = self.message_body

        domain = self.config.DRIVERS['mailgun']['domain']
        secret = self.config.DRIVERS['mailgun']['secret']
        return requests.post(
            "https://api.mailgun.net/v3/{0}/messages".format(domain),
            auth=("api", secret),
            data={
                "from": "{0} <mailgun@{1}>".format(self.config.FROM['name'], domain),
                "to": [self.to_address, "{0}".format(self.config.FROM['address'])],
                "subject": self.message_subject,
                "html": message})



""" A StartResponseProvider Service Provider """
from masonite.provider import ServiceProvider


class StartResponseProvider(ServiceProvider):

    def register(self):
        pass

    def boot(self, Request, Response, Headers):
        if not Request.redirect_url:
            # Convert the data that is retrieved above to bytes
            # so the wsgi server can handle it.

            data = bytes(Response, 'utf-8')

            self.app.bind('StatusCode', Request.get_status_code())
            Headers += [
                ("Content-Length", str(len(data)))
            ] + Request.get_cookies() + Request.get_headers()
        else:
            self.app.bind('StatusCode', "302 OK")
            self.app.bind('Headers', [
                ('Location', Request.redirect_url)
            ] + Request.get_cookies())

            Request.reset_redirections()

            self.app.bind('Response', 'redirecting ...')

        Request.reset_headers()
        Request.cookies = []
        if self.app.has('Session') and self.app.make('StatusCode') == '200 OK':
            self.app.make('Session').reset(flash_only=True)



import sys
import traceback
import os

from config import application

package_directory = os.path.dirname(os.path.realpath(__file__))


class ExceptionHandler:

    def __init__(self, app):
        self._app = app

        self._register_static_files()
    
    def _register_static_files(self):
        self._app.make('Storage').STATICFILES.update(
            {
                os.path.join(package_directory, 'snippets/exceptions'):
                '_exceptions/'
            }
        )
    
    def load_exception(self, exception):
        self._exception = exception

        self._render()
    
    def _render(self):
        self._app.bind('StatusCode', '500 Internal Server Error')

        # Run Any Framework Exception Hooks
        self._app.make('HookHandler').fire('*ExceptionHook')

        # Check if DEBUG is False
        if not self._app.make('Application').DEBUG or self._app.make('Application').DEBUG == 'False':
            raise self._exception

        # return a view
        rendered_view = self._app.make('View')('/masonite/snippets/exception',
            {
                'exception': self._exception,
                'traceback': traceback,
                'tb': sys.exc_info(),
                'app': self._app,
                'enumerate': enumerate,
                'open': open
            }
        ).rendered_template
        self._app.bind('Response', rendered_view)



''' A Module Description '''
from masonite.facades.Auth import Auth

class HomeController(object):
    ''' Home Dashboard Controller '''

    def __init__(self):
        pass

    def show(self, Request, Application):
        if not Auth(Request).user():
            Request.redirect('/login')
        return view('auth/home', {'app': Application, 'Auth': Auth(Request)})



""" Core of the IOC Container """

import inspect
from masonite.exceptions import MissingContainerBindingNotFound, ContainerError


class App():
    """
    Core of the Service Container. Performs bindings and resolving
    of objects to and from the container.
    """

    def __init__(self):
        self.providers = {}


    def bind(self, name, class_obj):
        """ 
        Bind classes into the container with a key value pair
        """
        self.providers.update({name: class_obj})
        return self


    def make(self, name):
        """
        Retreives a class from the container by key
        """
        if self.has(name):
            return self.providers[name]

        raise MissingContainerBindingNotFound("{0} key was not found in the container".format(name))


    def has(self, name):
        """
        Check if a key exists in the container
        """
        if name in self.providers:
            return True

        return False


    def helper(self):
        """
        Adds a helper to create builtin functions
        """
        return self


    def resolve(self, obj):
        """
        Takes an object such as a function or class method and resolves it's 
        parameters from objects in the container
        """
        provider_list = []

        for parameter, value in inspect.signature(obj).parameters.items():
            if ':' in str(value):
                provider_list.append(self._find_annotated_parameter(value))
            else:
                provider_list.append(self._find_parameter(value))

        return obj(*provider_list)

    def collect(self, search):
        # Search Can Be:
        #    '*ExceptionHook'
        #    'Sentry*'
        #    'Sentry*Hook'
        provider_list = {}
        for key, value in self.providers.items():
            if search.startswith('*'):
                if key.endswith(search.split('*')[1]):
                    provider_list.update({key: value})
            elif search.endswith('*'):
                if key.startswith(search.split('*')[0]):
                    provider_list.update({key: value})
            elif '*' in search:
                split_search = search.split('*')
                print(split_search)
                print('look for something that starts with ', split_search[0])
                print('and ends with ', split_search[1])
                if key.startswith(split_search[0]) and key.endswith(split_search[1]):
                    print('found a match for ', split_search[0], ' and ending with ', split_search[1])
                    provider_list.update({key: value})
            else:
                raise AttributeError("There is no '*' in your collection search")
        return provider_list

    def _find_parameter(self, parameter):
        """
        Find a parameter in the container
        """
        parameter = str(parameter)
        if parameter is not 'self' and parameter in self.providers:
            return self.providers[parameter]
        
        raise ContainerError(
            'The dependency with the key of {0} could not be found in the container'.format(parameter)
        )


    def _find_annotated_parameter(self, parameter):
        """
        Find a given annotation in the container
        """
        for provider, provider_class in self.providers.items():
            if parameter.annotation == provider_class.__class__ or parameter.annotation == provider_class or isinstance(provider_class, parameter.annotation.__class__):
                return provider_class
        
        raise ContainerError('The dependency with the {0} annotation could not be resolved by the container'.format(parameter))



import binascii
import os


class Csrf(object):
    """
    Class for csrf protection
    """

    def __init__(self, request):
        self.request = request

    def generate_csrf_token(self):
        """
        Generate token for csrf protection
        """

        token = bytes(binascii.b2a_hex(os.urandom(15))).decode('utf-8')
        self.request.cookie('csrf_token', token, encrypt=False)
        return token

    def verify_csrf_token(self, token):
        """
        Verify if csrf token is valid
        """

        if self.request.get_cookie('csrf_token', decrypt=False) == token:
            return True
        else:
            return False



import os
from cleo import Command


class ModelCommand(Command):
    """
    Creates a model

    model
        {name : Name of the model}
    """

    def handle(self):
        model = self.argument('name')
        if not os.path.isfile('app/{0}.py'.format(model)):
            if not os.path.exists(os.path.dirname('app/{0}.py'.format(model))):
                # Create the path to the model if it does not exist
                os.makedirs(os.path.dirname('app/{0}.py'.format(model)))

            f = open('app/{0}.py'.format(model), 'w+')

            f.write("''' A {0} Database Model '''\n".format(model))
            f.write('from config.database import Model\n\n')
            f.write("class {0}(Model):\n    pass\n".format(model))

            self.info('Model Created Successfully!')
        else:
            self.error('Model Already Exists!')



import os
from cleo import Command


class ViewCommand(Command):
    """
    Creates a view

    view
        {name : Name of the view you would like to create}
    """

    def handle(self):
        name = self.argument('name')
        if os.path.isfile('resources/templates/' + name + '.html'):
            self.error('{0} View Exists!'.format(name))
        else:
            open('resources/templates/' + name + '.html', 'w+')
            self.info('{0} View Created Successfully!'.format(name))



import os
import shutil
from subprocess import call

from cleo import Command


class InstallCommand(Command):
    """
    Installs all of Masonite's dependencies

    install
    """

    def handle(self):
        call(["pip3", "install", "-r", "requirements.txt"])

        # create the .env file if it does not exist
        if not os.path.isfile('.env'):
            shutil.copy('.env-example', '.env')

        call(["craft", "key", "--store"])



from masonite.exceptions import DriverNotFound, MissingContainerBindingNotFound


class Manager:
    """
    Manager container class
    """

    config = None
    driver_prefix = None

    def __init__(self, container=None):
        self.manage_driver = None
        self.container = container

    def load_container(self, container):
        self.container = container
        self.create_driver()
        return self

    def driver(self, driver):
        self.create_driver(driver)
        return self.container.resolve(self.manage_driver).load_manager(self)

    def create_driver(self, driver=None):
        if not driver:
            driver = self.container.make(self.config).DRIVER.capitalize()
        else:
            driver = driver.capitalize()

        try:
            self.manage_driver = self.container.make(
                '{0}{1}Driver'.format(self.driver_prefix, driver)
            )
        except MissingContainerBindingNotFound:
            raise DriverNotFound(
                'Could not find the {0}{1}Driver from the service container. Are you missing a service provider?'.format(self.driver_prefix, driver))



from masonite.exceptions import FileTypeException
from masonite.drivers.BaseDriver import BaseDriver


class BaseUploadDriver(BaseDriver):
    """
    Class base for upload drivers
    """

    # this will accept all file types
    accept_file_types = None

    def accept(self, *args, **kwargs):
        """
        Set file types to accept
        """

        self.accept_file_types = args
        return self

    def validate_extension(self, filename):
        """
        Check if files is a valid extension
        """

        if self.accept_file_types is not None:
            if not filename.endswith(self.accept_file_types):
                raise FileTypeException("The extension file not is valid.")



from masonite.routes import Get, Post, Delete, Patch, Put


def flatten_routes(routes):
    """
    Flattens the grouped routes into a single list of routes.
    """
    route_collection = []
    for route in routes:
        # Check if a route is a list of routes
        if isinstance(route, list):
            for r in flatten_routes(route):
                route_collection.append(r)
        else:
            route_collection.append(route)

    return route_collection


def get(url, controller):
    """
    Shortcut for GET http
    """
    return Get().route(url, controller)


def post(url, controller):
    """
    Shortcut for POST http
    """
    return Post().route(url, controller)


def delete(url, controller):
    """
    Shortcut for DELETE http
    """
    return Delete().route(url, controller)


def put(url, controller):
    """
    Shortcut for Put http
    """
    return Put().route(url, controller)


def patch(url, controller):
    """
    Shortcut for PATCH http
    """
    return Patch().route(url, controller)


def group(url, route_list):
    """
    Group route
    """
    for route in route_list:
        route.route_url = url + route.route_url

    return route_list



from masonite.contracts.BroadcastContract import BroadcastContract
from masonite.exceptions import DriverLibraryNotFound
from masonite.drivers.BaseDriver import BaseDriver


class BroadcastPusherDriver(BroadcastContract, BaseDriver):

    def __init__(self, BroadcastConfig):
        self.config = BroadcastConfig
        self.ssl_message = True


    def ssl(self, boolean):
        self.ssl_message = boolean
        return self

    def channel(self, channels, message, event='base-event'):
        try:
            import pusher
        except ImportError:
            raise DriverLibraryNotFound(
                'Could not find the "pusher" library. Please pip install this library running "pip install pusher"')

        pusher_client = pusher.Pusher(
            app_id=self.config.DRIVERS['pusher']['app_id'],
            key=self.config.DRIVERS['pusher']['client'],
            secret=self.config.DRIVERS['pusher']['secret'],
            ssl=self.ssl_message
        )

        if isinstance(message, str):
            message = {'message': message}  

        if isinstance(channels, list):
            for channel in channels:
                pusher_client.trigger(channel, event, message)
        else:
            pusher_client.trigger(channels, event, message)

        return message



""" A Cache Service Provider """
from masonite.provider import ServiceProvider
from masonite.managers.CacheManager import CacheManager
from masonite.drivers.CacheDiskDriver import CacheDiskDriver
from config import cache


class CacheProvider(ServiceProvider):

    wsgi = False

    def register(self):
        self.app.bind('CacheConfig', cache)
        self.app.bind('CacheDiskDriver', CacheDiskDriver)
        self.app.bind('CacheManager', CacheManager(self.app))

    def boot(self, CacheManager, CacheConfig):
        self.app.bind('Cache', CacheManager.driver(CacheConfig.DRIVER))



from abc import ABC, abstractmethod


class CacheContract(ABC):

    @abstractmethod
    def store(self): pass

    @abstractmethod
    def store_for(self): pass

    @abstractmethod
    def get(self): pass

    @abstractmethod
    def is_valid(self): pass

    @abstractmethod
    def cache_exists(self): pass



""" Controller module """


class Controller:
    """
    Base Class for Controllers
    """

    def __init__(self):
        self.app = None

    def load_app(self, app):
        """ Loads the container into the controller """
        self.app = app
        return self



""" A RedirectionProvider Service Provider """
from config import session
from masonite.drivers.SessionMemoryDriver import SessionMemoryDriver
from masonite.drivers.SessionCookieDriver import SessionCookieDriver
from masonite.managers.SessionManager import SessionManager
from masonite.provider import ServiceProvider


class SessionProvider(ServiceProvider):

    def register(self):
        self.app.bind('SessionConfig', session)
        self.app.bind('SessionMemoryDriver', SessionMemoryDriver)
        self.app.bind('SessionCookieDriver', SessionCookieDriver)
        self.app.bind('SessionManager', SessionManager(self.app))

    def boot(self, Environ, Request, ViewClass, SessionManager, SessionConfig):
        self.app.bind('Session', SessionManager.driver(SessionConfig.DRIVER))
        Session = self.app.make('Session')
        Request.session = Session

        ViewClass.share({
            'session': Session.helper
        })




from masonite.contracts.UploadContract import UploadContract
from masonite.drivers.BaseUploadDriver import BaseUploadDriver
from masonite.exceptions import DriverLibraryNotFound


class UploadS3Driver(BaseUploadDriver, UploadContract):
    """
    Amazon S3 Upload driver
    """

    def __init__(self, UploadManager, StorageConfig):
        self.upload = UploadManager
        self.config = StorageConfig

    def store(self, fileitem, location=None):
        file_location = self.upload.driver('disk').store(fileitem)

        # Check if is a valid extension
        self.validate_extension(fileitem.filename)

        try:
            import boto3
        except ImportError:
            raise DriverLibraryNotFound(
                'Could not find the "ably" library. Please pip install this library running "pip install ably"')

        session = boto3.Session(
            aws_access_key_id=self.config.DRIVERS['s3']['client'],
            aws_secret_access_key=self.config.DRIVERS['s3']['secret'],
        )

        s3 = session.resource('s3')

        s3.meta.client.upload_file(
            file_location,
            self.config.DRIVERS['s3']['bucket'],
            fileitem.filename
        )

    def store_prepend(self, fileitem, prepend, location=None):
        fileitem.filename = prepend + fileitem.filename

        return self.store(self, fileitem, location)



from masonite.app import App
from masonite.testsuite.TestRoute import TestRoute
from masonite.testsuite.TestRequest import TestRequest
from config import application
from pydoc import locate


WSGI_REQUEST = {
    'wsgi.version': (1, 0),
    'wsgi.multithread': False,
    'wsgi.multiprocess': True,
    'wsgi.run_once': False,
    'SERVER_SOFTWARE': 'gunicorn/19.7.1',
    'REQUEST_METHOD': 'GET',
    'QUERY_STRING': 'application=Masonite',
    'RAW_URI': '/',
    'SERVER_PROTOCOL': 'HTTP/1.1',
    'HTTP_HOST': '127.0.0.1:8000',
    'HTTP_ACCEPT': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
    'HTTP_UPGRADE_INSECURE_REQUESTS': '1',
    'HTTP_COOKIE': 'setcookie=value',
    'HTTP_USER_AGENT': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_2) AppleWebKit/604.4.7 (KHTML, like Gecko) Version/11.0.2 Safari/604.4.7',
    'HTTP_ACCEPT_LANGUAGE': 'en-us',
    'HTTP_ACCEPT_ENCODING': 'gzip, deflate',
    'HTTP_CONNECTION': 'keep-alive',
    'wsgi.url_scheme': 'http',
    'REMOTE_ADDR': '127.0.0.1',
    'REMOTE_PORT': '62241',
    'SERVER_NAME': '127.0.0.1',
    'SERVER_PORT': '8000',
    'PATH_INFO': '/',
    'SCRIPT_NAME': ''
}


class TestSuite:

    def create_container(self):
        container = App()

        container.bind('WSGI', object)
        container.bind('Application', application)

        """
        |--------------------------------------------------------------------------
        | Bind all service providers
        |--------------------------------------------------------------------------
        |
        | Let's register everything into the Service Container. Once everything is
        | in the container we can run through all the boot methods. For reasons
        | some providers don't need to execute with every request and should
        | only run once when the server is started. Providers will be ran
        | once if the wsgi attribute on a provider is False.
        |
        """

        for provider in container.make('Application').PROVIDERS:
            locate(provider)().load_app(container).register()

        for provider in container.make('Application').PROVIDERS:
            located_provider = locate(provider)().load_app(container)

            if located_provider.wsgi is False:
                container.resolve(locate(provider)().load_app(container).boot)

            """
        |--------------------------------------------------------------------------
        | Startup the Service Container
        |--------------------------------------------------------------------------
        |
        | Instantiate the Service Container so we can bind classes into it and 
        | bind the environ variable that is created by the WSGI server into
        | the container.
        |
        """

        container.bind('Environ', WSGI_REQUEST)

        """
        |--------------------------------------------------------------------------
        | Execute All Service Providers
        |--------------------------------------------------------------------------
        |
        | Run all service provider boot methods if the wsgi attribute is true.
        |
        """

        for provider in container.make('Application').PROVIDERS:
            located_provider = locate(provider)().load_app(container)
            container.bind('Response', 'test')
            if located_provider.wsgi is True:
                container.resolve(located_provider.boot)

        self.container = container
        return self

    def get_container(self):
        return self.container

    def route(self, route):
        return TestRoute(route)

    def get(self, route_url):
        return TestRequest().get(route_url)


from config import storage, application
import os


class Storage:

    def __init__(self):
        pass

    # this function will compile sass files only if
    # the libsass module installed
    def compile_sass(self):
        try:
            import sass
        except ImportError as e:
            pass
        else:
            matches = []
            for files in storage.SASSFILES['importFrom']:
                for root, dirnames, filenames in os.walk(os.path.join(application.BASE_DIRECTORY, files)):
                    for filename in filenames:
                        if filename.endswith(('.sass', '.scss')) and not filename.startswith('_'):
                            matches.append(os.path.join(root, filename))

            for filename in matches:
                with open(filename) as f:
                    compiled_sass = sass.compile(
                        string=f.read(), include_paths=storage.SASSFILES['includePaths']
                    )
                    name = filename.split(os.sep)[-1].replace('.scss', '').replace('.sass', '')
                    write_file = os.path.join(os.path.join(application.BASE_DIRECTORY, 
                        storage.SASSFILES['compileTo']), '{0}.css'.format(name))
                with open(write_file, 'w') as r:
                    r.write(compiled_sass)



from setuptools import setup

setup(
    name='masonite',
    packages=[
        'masonite',
        'masonite.auth',
        'masonite.facades',
        'masonite.providers',
        'masonite.commands',
        'masonite.drivers',
        'masonite.managers',
        'masonite.testsuite',
        'masonite.queues',
        'masonite.contracts',
        'masonite.helpers',
    ],
    version='1.6.6',
    install_requires=[
        'validator.py==1.2.5',
        'cryptography==2.2.2',
        'bcrypt==3.1.4',
        'requests==2.18.4',
        'tldextract==2.2.0',
        'Jinja2==2.10',
        'python-dotenv==0.7.1',
        'passlib==1.7.1',
        'whitenoise==3.3.1',
        'bcrypt==3.1.4',
        'pytest==3.3.1',
        'orator==0.9.7',
        'cryptography==2.1.4',
        'PyMySQL==0.8.0',
        'psycopg2==2.7.3.2',
        'masonite-entry>=0.0.0,<=0.9.99',
        'pendulum==1.4.4',
        'cleo==0.6.5',
    ],
    description='The core for the Masonite framework',
    author='Joseph Mancuso',
    author_email='idmann509@gmail.com',
    url='https://github.com/MasoniteFramework/masonite',
    keywords=['python web framework', 'python3'],
    classifiers=[],
    include_package_data=True,
)



from masonite.contracts.SessionContract import SessionContract
from masonite.drivers.BaseDriver import BaseDriver


class SessionMemoryDriver(SessionContract, BaseDriver):
    """
    Session from the memory driver
    """

    _session = {}
    _flash = {}

    def __init__(self, Environ):
        """
        Constructor
        """

        self.environ = Environ

    def get(self, key):
        """
        Get a session from object _session
        """

        data = self.__collect_data(key)
        if data:
            return data

        return None

    def set(self, key, value):
        """
        Set a new session in object _session
        """

        ip = self.__get_client_address()

        if ip not in self._session:
            self._session[ip] = {}

        self._session[ip][key] = value

    def has(self, key):
        """
        Check if a key exists in the session
        """

        data = self.__collect_data()
        if data and key in data:
            return True
        return False

    def all(self):
        """
        Get all session data
        """

        return self.__collect_data()

    def flash(self, key, value):
        """
        Add temporary data to the session
        """

        ip = self.__get_client_address()
        if ip not in self._flash:
            self._flash[ip] = {}

        self._flash[ip][key] = value

    def reset(self, flash_only=False):
        """
        Reset object _session
        """

        ip = self.__get_client_address()

        if flash_only:
            if ip in self._flash:
                self._flash[ip] = {}
        else:
            if ip in self._session:
                self._session[ip] = {}

    def __get_client_address(self):
        """
        Get ip from the client
        """

        if 'HTTP_X_FORWARDED_FOR' in self.environ:
            return self.environ['HTTP_X_FORWARDED_FOR'].split(',')[-1].strip()

        return self.environ['REMOTE_ADDR']

    def __collect_data(self, key=False):
        """
        Collect data from session and flash data
        """

        ip = self.__get_client_address()

        # Declare a new dictionary
        session = {}

        # If the session data has keys
        if ip in self._session:
            session = self._session[ip]

        # If the session flash has keys
        if ip in self._flash:
            session.update(self._flash[ip])

        # If a key is set and it is inside the new declared session, return that key
        if key and key in session:
            return session[key]

        # If the key is set and is not in the session
        if key and key not in session:
            return None

        # If the session is still an empty dictionary
        if not session:
            return None
        
        # No checks have been hit. Return the new dictionary
        return session

    def helper(self):
        """
        Used to create builtin helper function
        """

        return self



from abc import ABC, abstractmethod


class MailContract(ABC):

    @abstractmethod
    def to(self): pass

    @abstractmethod
    def template(self): pass

    @abstractmethod
    def send_from(self): pass

    @abstractmethod
    def subject(self): pass



''' A Module Description '''
from masonite.facades.Auth import Auth
from config import auth
import bcrypt


class RegisterController(object):
    ''' Class Docstring Description '''

    def __init__(self):
        pass

    def show(self, Request, Application):
        ''' Show the registration page '''
        return view('auth/register', {'app': Application, 'Auth': Auth(Request)})

    def store(self, Request):
        ''' Register a new user '''
        # register the user
        password = bytes(bcrypt.hashpw(
            bytes(Request.input('password'), 'utf-8'), bcrypt.gensalt()
        )).decode('utf-8')

        auth.AUTH['model'].create(
            name=Request.input('name'),
            password=password,
            email=Request.input('email'),
        )

        # login the user
        # redirect to the homepage
        if Auth(Request).login(Request.input(auth.AUTH['model'].__auth__), Request.input('password')):
            return Request.redirect('/home')

        return Request.redirect('/register')



import os
import sys
from subprocess import check_output

from cleo import Command

from masonite.packages import add_venv_site_packages


class MigrateCommand(Command):
    """
    Run migrations

    migrate
    """

    def handle(self):
        sys.path.append(os.getcwd())
        try:
            add_venv_site_packages()
        except ImportError:
            self.comment('This command must be ran inside of the root of a Masonite project directory')

        from wsgi import container

        migration_directory = ['databases/migrations']
        for key, value in container.providers.items():
            if 'MigrationDirectory' in key:
                migration_directory.append(value)

        for directory in migration_directory:
            self.line('')
            if len(migration_directory) > 1:
                self.info('Migrating: {0}'.format(directory))
            try:
                output = bytes(check_output(
                        ['orator', 'migrate', '-c', 'config/database.py', '-p', directory, '-f']
                    )).decode('utf-8')

                self.line(
                    output.replace('OK', '<info>OK</info>') \
                    .replace('Migrated', '<info>Migrated</info><fg=cyan>') + '</>'
                )
            except Exception:
                pass



""" An AppProvider Service Provider """
from masonite.commands.AuthCommand import AuthCommand
from masonite.commands.CommandCommand import CommandCommand
from masonite.commands.ControllerCommand import ControllerCommand
from masonite.commands.InstallCommand import InstallCommand
from masonite.commands.JobCommand import JobCommand
from masonite.commands.KeyCommand import KeyCommand
from masonite.commands.MakeMigrationCommand import MakeMigrationCommand
from masonite.commands.MigrateCommand import MigrateCommand
from masonite.commands.MigrateRefreshCommand import MigrateRefreshCommand
from masonite.commands.MigrateResetCommand import MigrateResetCommand
from masonite.commands.MigrateRollbackCommand import MigrateRollbackCommand
from masonite.commands.ModelCommand import ModelCommand
from masonite.commands.ProviderCommand import ProviderCommand
from masonite.commands.ServeCommand import ServeCommand
from masonite.commands.ViewCommand import ViewCommand
from masonite.exception_handler import ExceptionHandler
from masonite.helpers.routes import flatten_routes
from masonite.hook import Hook
from masonite.provider import ServiceProvider
from masonite.request import Request
from masonite.routes import Route

from config import storage
from routes import api, web


class AppProvider(ServiceProvider):

    def register(self):
        self.app.bind('HookHandler', Hook(self.app))
        self.app.bind('WebRoutes', flatten_routes(web.ROUTES))
        self.app.bind('Response', None)
        self.app.bind('Storage', storage)
        self.app.bind('Route', Route())
        self.app.bind('Request', Request())
        self.app.bind('Container', self.app)
        self.app.bind('ExceptionHandler', ExceptionHandler(self.app))

        # Insert Commands
        self.app.bind('MasoniteAuthCommand', AuthCommand())
        self.app.bind('MasoniteCommandCommand', CommandCommand())
        self.app.bind('MasoniteControllerCommand', ControllerCommand())
        self.app.bind('MasoniteInstallCommand', InstallCommand())
        self.app.bind('MasoniteJobCommand', JobCommand())
        self.app.bind('MasoniteKeyCommand', KeyCommand())
        self.app.bind('MasoniteMakeMigrationCommand', MakeMigrationCommand())
        self.app.bind('MasoniteMigrateCommand', MigrateCommand())
        self.app.bind('MasoniteMigrateRefreshCommand', MigrateRefreshCommand())
        self.app.bind('MasoniteMigrateResetCommand', MigrateResetCommand())
        self.app.bind('MasoniteMigrateRollbackCommand', MigrateRollbackCommand())
        self.app.bind('MasoniteModelCommand', ModelCommand())
        self.app.bind('MasoniteProviderCommand', ProviderCommand())
        self.app.bind('MasoniteViewCommand', ViewCommand())
        self.app.bind('MasoniteServeCommand', ServeCommand())

    def boot(self, Environ, Request, Route):
        self.app.bind('Headers', [])
        Route.load_environ(Environ)
        Request.load_environ(Environ).load_app(self.app)



from masonite.contracts.SessionContract import SessionContract
from masonite.drivers.BaseDriver import BaseDriver


class SessionCookieDriver(SessionContract, BaseDriver):
    """
    Session from the memory driver
    """

    def __init__(self, Environ, Request):
        """
        Constructor
        """

        self.environ = Environ
        self.request = Request

    def get(self, key):
        """
        Get a session from object _session
        """

        cookie = self.request.get_cookie('s_{0}'.format(key))
        if cookie:
            return cookie

        return None

    def set(self, key, value):
        """
        Set a new session in object _session
        """

        self.request.cookie('s_{0}'.format(key), value)

    def has(self, key):
        """
        Check if a key exists in the session
        """

        if self.get(key):
            return True
        return False

    def all(self):
        """
        Get all session data
        """

        return self.__collect_data()

    def __collect_data(self):
        """
        Collect data from session and flash data
        """

        cookies = {}
        if 'HTTP_COOKIE' in self.environ and self.environ['HTTP_COOKIE']:
            cookies_original = self.environ['HTTP_COOKIE'].split(';')
            for cookie in cookies_original:
                if cookie.startswith('s_'):
                    data = cookie.split("=")
                    cookie_value = self.request.get_cookie(data[0])
                    cookies[data[0][2:]] = cookie_value
        return cookies

    def flash(self, key, value):
        """
        Add temporary data to the session
        """

        self.request.cookie('s_{0}'.format(key), value, expires='2 seconds')

    def reset(self, flash_only=False):
        """
        Reset object _session
        """
        cookies = self.__collect_data()
        for cookie in cookies:
            self.request.delete_cookie('s_{0}'.format(cookie))

    def helper(self):
        """
        Used to create builtin helper function
        """

        return self
